#ifndef POLYNOME_HPP
#define POLYNOME_HPP
#include <cmath>
#include <iostream>
#include "complexe.hpp"
using namespace std;

class segment {
public:
  static constexpr double a = 0.0;
  static constexpr double b = 1.0;
};

class segment_ref {
public:
  static constexpr double a = -1.0;
  static constexpr double b = 1.0;
};
  
  
template <class segment_type=segment_ref> class polynome {
  int deg;
  complexe * coefs;
public: 
  polynome (){
    deg = 0;
    coefs = NULL;
  }
  polynome (int degree){
    deg = degree;
    coefs = new complexe[degree+1];
    for(int i =0; i<deg+1;i++){
	coefs[i] = complexe(0,0);
    }
  }
  polynome (const polynome & x){
    deg = x.deg;
    coefs = NULL;
    if(x.coefs != NULL){
      coefs = new complexe[deg+1];
      for(int i =0; i<deg+1;i++){
  	coefs[i] = x.coefs[i];
      }
    }
  }
  ~polynome (){
    if(coefs != NULL){
      delete [] coefs;
    }
  }
  complexe get(int);
  void set(complexe, int);
  void adresse();
  polynome primitive();
  polynome derivative();
  complexe integral();
  polynome conjuge();
  polynome operator + (polynome);
  polynome operator - (polynome);
  polynome operator * (polynome);
  polynome operator / (polynome);
  complexe operator , (polynome);
  complexe operator () (complexe);
  complexe operator [] (int i);
  polynome & operator = (const polynome &);
  complexe norm2();
  bool operator <(polynome);
  bool operator >(polynome);
  bool operator ==(polynome);
  template <class type>
  friend ostream & operator<<(ostream & out, const polynome<type> &);
  template <class type>
  friend istream & operator>>(istream & in, polynome<type> &);
  int testu_1();
  int testu_2();
  int testu_3();
  int all_testsu();

private:
  polynome convert(int deg2){
    if( deg2 < deg){
      cout<<"convertion pas possible"<<endl;
    }
    polynome res(deg2);
    for(int i=0;i<deg+1;i++){
	res.coefs[i] = coefs[i];
    }
    return res;   
  }
  
};		   



template <class segment_type> void polynome<segment_type>::adresse(){
  cout<<"adresse :"<<coefs<<endl;
};

template <class segment_type> complexe polynome<segment_type>:: get(int i){
  complexe res;
  res= coefs[i];
  return res;
};

template <class segment_type> void polynome<segment_type>:: set(complexe x, int i){
  coefs[i] = x;
};

template <class segment_type> polynome<segment_type>  polynome<segment_type>:: operator + (polynome p){
  int degt;
  degt = max(deg,p.deg);
  polynome<segment_type> res(degt);
  polynome<segment_type> temp;
  
  if(deg < p.deg){ temp = this->convert(p.deg);
    for(int i=0;i<degt+1;i++){
      res.coefs[i] = temp.coefs[i] + p.coefs[i];
    }
  }
  if(deg > p.deg){ temp = p.convert(deg); 
    for(int i =0;i<degt+1;i++){
      res.coefs[i]= coefs[i] + temp.coefs[i];
    }
  }
  if(deg == p.deg){ 
    for(int i =0;i<degt+1;i++){
      res.coefs[i]= coefs[i] + p.coefs[i];  }
  }
  return res;
}  

template <class segment_type> polynome<segment_type>  polynome<segment_type>:: operator - (polynome<segment_type> p){
  int degt;
  degt = max(deg,p.deg);
  polynome<segment_type> res(degt);
  polynome<segment_type> temp;
  
  if(deg < p.deg){ temp = this->convert(p.deg);
    for(int i=0;i<degt+1;i++){
      res.coefs[i] = temp.coefs[i] - p.coefs[i];
    }
  }
  if(deg > p.deg){ temp = p.convert(deg); 
    for(int i =0;i<degt+1;i++){
      res.coefs[i]= coefs[i] - temp.coefs[i];
    }
  }
  if(deg == p.deg){ 
    for(int i =0;i<degt+1;i++){
      res.coefs[i]= coefs[i] - p.coefs[i];  }
  }
  return res;
}

template <class segment_type> polynome<segment_type> polynome<segment_type>:: operator * (polynome<segment_type> p){
  int degt=deg+p.deg;
  polynome<segment_type> res(degt);
  polynome<segment_type> temp, temp2;

  temp = this->convert(degt);
  temp2 = p.convert(degt);
  for(int i=0;i<degt+1;i++){
    for(int j=0;j<i+1;j++){
      res.coefs[i]=res.coefs[i]+temp.coefs[j]*temp2.coefs[i-j];
    }
  }
  return res;
} 

template <class segment_type> polynome<segment_type> & polynome<segment_type>::operator =(const polynome<segment_type> & p){
  if(&p != this){
    deg = p.deg;
    if(coefs != NULL){ delete [] coefs;}
    coefs = NULL;
    if(p.coefs != NULL){
      coefs = new complexe[deg+1];
      for(int i =0;i<deg+1;i++){
  	coefs[i]=p.coefs[i];
      }
    }
  }
  return *this;
}

template <class segment_type> complexe polynome<segment_type>::operator [](int i){
  return coefs[i];
}

template <class segment_type> complexe polynome<segment_type>::operator ()(complexe x){
  complexe res(0.0,0.0);
  complexe prod(0.0,0.0);
  for(int i =0;i<deg+1;i++){
    prod =complexe(1.0,0.0);
    for(int j=0;j<i;j++){ prod = prod*x; }
      res = res + coefs[i]*prod;
  }
  return res;
}

template <class segment_type> complexe polynome<segment_type>:: operator,(polynome p){
  polynome pro, prim;
  complexe res(0.0,0.0);

  pro = (*this)*p.conjuge();
  res = pro.integral();
  return res;
}

template <class segment_type> complexe polynome<segment_type>::norm2(){
  complexe n(0,0);
  n = (*this,*this);
  return n;
};

template <class segment_type> bool polynome<segment_type>:: operator < (polynome<segment_type> x){
  bool res = false;
  if(this->norm2() < x.norm2() ) { res = true;}
  return res;
};

template <class segment_type> bool polynome<segment_type>:: operator > (polynome<segment_type> x){
  bool res = false;
  if(this->norm2() > x.norm2()) { res = true;}
  return res;
};

template <class segment_type> bool polynome<segment_type>:: operator == (polynome<segment_type> x){
  bool res = false;
  if(this->norm2() == x.norm2()) { res = true;}
  return res;
};



template <class segment_type> polynome<segment_type> polynome<segment_type>::conjuge(){
  polynome<segment_type> res(deg);
  for(int i=0;i<deg+1;i++){
    res.coefs[i]=coefs[i].conjuge();
  }
  return res;
}

template <class segment_type> polynome<segment_type> polynome<segment_type>::derivative(){
  polynome<segment_type> res(deg-1);
  for(int i =1;i<deg+1;i++){
    res.coefs[i-1]= i*coefs[i];
  }
  return res;
};

template <class segment_type> polynome<segment_type> polynome<segment_type>::primitive(){
  polynome<segment_type> res(deg+1);
  for(int i=0;i<deg+1;i++){
    res.coefs[i+1]= (1.0/(i+1.0))*coefs[i];
  }
  res.coefs[0]=complexe(0.0,0.0);
  return res;
}

template <class segment_type> complexe polynome<segment_type>::integral(){
  complexe res(0.0,0.0);
  //polynome<segment_type> primitive;
  auto primitive= (*this);
  primitive = this->primitive();
  res= primitive(complexe(segment_type::b,0.0))-primitive(complexe(segment_type::a,0.0));
  return res;
}

template <class segment_type> ostream & operator << (ostream & out,const polynome<segment_type> & p){
  for(int i=0;i<p.deg+1;i++){
    out<<"("<<p.coefs[i]<<")"<<"x^"<<i;
    if(i!= p.deg){ out<<" + ";}
  }
  return out;
}

template <class segment_type> istream & operator >>(istream & in, polynome<segment_type> & p){
  for(int i=0;i<p.deg+1;i++){
    cout<<" coef: "<<i<<endl;
    in >> p.coefs[i];
  }
  return in;
}

template <class segment_type> int polynome<segment_type>:: testu_1(){
  polynome<segment> P(2);
  polynome<segment> Q,R,S;
  complexe x(1.0,1.0), y(2.0,-1.0), z(1.0,0.0);

  *this = P;
  this->set(x,0); this->set(y,1); this->set(z,2);
  Q = this->derivative();
  R = this->primitive();
  S = *this+R;
  *this = S*Q;
  
  complexe x0(3.0,1.0), y0(8.0,-1.0);
  if((*this)[0] == x0 && (*this)[1] == y0){ return 0;}
  else { return 1;}
};

template <class segment_type> int polynome<segment_type>:: testu_2(){
  polynome<segment> P(1), Q,S;
  complexe x(1.0,0.0), y(2.0,0.0);
  complexe ps, ps0(4.0,0.0);

  P.set(x,0); P.set(y,1);
  *this = P;
  Q = this->derivative();
  ps=(*this,Q);
  if(ps == ps0){ return 0;}
  else { return 1;}
  
  return 0;
};

template <class segment_type> int polynome<segment_type>:: testu_3(){
  polynome<segment> P(1);
  polynome<segment> Q,S;
  complexe x(1.0,-1.0), y(2.0,0.0);
  bool a=false,b=false,c=false;

  P.set(x,0); P.set(y,1);
  S = P;
  *this = P;
  Q = this->derivative();
  a = (*this > Q);
  b = (Q < *this);
  c = (S == *this);
  if( a+b+c == 3) { return 0; }
  else { return 1; }

  return 0; 
}

template <class segment_type> int polynome<segment_type>:: all_testsu(){
  int c=0,t=0;
  polynome<segment> p;
  c=p.testu_1(); t = t +c;
  c=p.testu_2(); t = t +c;
  c=p.testu_3(); t = t +c;
  cout<<"Tests unitaires réussis, classe polynome: "<<3-t<<"/3"<<endl;
  return 0;
}

#endif
